@@ -6,7 +6,7 @@
 
 /* Standard definitions */
 #include "Python.h"
-#include <stddef.h>
+#include <setjmp.h>
 #include <signal.h>
 #include <errno.h>
 #include <sys/time.h>
@@ -28,8 +28,11 @@
 
 /* GNU readline definitions */
 #undef HAVE_CONFIG_H /* Else readline/chardefs.h includes strings.h */
-#include <readline/readline.h>
-#include <readline/history.h>
+/* Ensure that we get the intended local checkout of GNU readline
+ * and not some system version (which could be libedit in disguise).
+ */
+#include "../../readline/readline.h"
+#include "../../readline/history.h"
 
 #ifdef HAVE_RL_COMPLETION_MATCHES
 #define completion_matches(x, y) \
@@ -45,14 +48,14 @@ extern char **completion_matches(char *, CPFunction *);
 #endif
 #endif
 
+#ifdef __APPLE__
 /*
  * It is possible to link the readline module to the readline
  * emulation library of editline/libedit.
  *
- * This emulation library is not 100% API compatible with the "real" readline
- * and cannot be detected at compile-time,
- * hence we use a runtime check to detect if the Python readlinke module is
- * linked to libedit.
+ * On OSX this emulation library is not 100% API compatible
+ * with the "real" readline and cannot be detected at compile-time,
+ * hence we use a runtime check to detect if we're using libedit
  *
  * Currently there is one known API incompatibility:
  * - 'get_history' has a 1-based index with GNU readline, and a 0-based
@@ -64,6 +67,7 @@ static int using_libedit_emulation = 0;
 static const char libedit_version_tag[] = "EditLine wrapper";
 
 static int libedit_history_start = 0;
+#endif /* __APPLE__ */
 
 #ifdef HAVE_RL_COMPLETION_DISPLAY_MATCHES_HOOK
 static void
@@ -75,119 +79,28 @@ on_completion_display_matches_hook(char **matches,
    (see issue #17289 for the motivation). */
 static char *completer_word_break_characters;
 
-typedef struct {
-  /* Specify hook functions in Python */
-  PyObject *completion_display_matches_hook;
-  PyObject *startup_hook;
-  PyObject *pre_input_hook;
-
-  PyObject *completer; /* Specify a word completer in Python */
-  PyObject *begidx;
-  PyObject *endidx;
-} readlinestate;
-
-
-#define readline_state(o) ((readlinestate *)PyModule_GetState(o))
-
-static int
-readline_clear(PyObject *m)
-{
-   readlinestate *state = readline_state(m);
-   Py_CLEAR(state->completion_display_matches_hook);
-   Py_CLEAR(state->startup_hook);
-   Py_CLEAR(state->pre_input_hook);
-   Py_CLEAR(state->completer);
-   Py_CLEAR(state->begidx);
-   Py_CLEAR(state->endidx);
-   return 0;
-}
-
-static int
-readline_traverse(PyObject *m, visitproc visit, void *arg)
-{
-    readlinestate *state = readline_state(m);
-    Py_VISIT(state->completion_display_matches_hook);
-    Py_VISIT(state->startup_hook);
-    Py_VISIT(state->pre_input_hook);
-    Py_VISIT(state->completer);
-    Py_VISIT(state->begidx);
-    Py_VISIT(state->endidx);
-    return 0;
-}
-
-static void
-readline_free(void *m)
-{
-    readline_clear((PyObject *)m);
-}
-
-static PyModuleDef readlinemodule;
-
-#define readlinestate_global ((readlinestate *)PyModule_GetState(PyState_FindModule(&readlinemodule)))
-
-
-/* Convert to/from multibyte C strings */
-
-static PyObject *
-encode(PyObject *b)
-{
-    return PyUnicode_EncodeLocale(b, "surrogateescape");
-}
-
-static PyObject *
-decode(const char *s)
-{
-    return PyUnicode_DecodeLocale(s, "surrogateescape");
-}
-
-
-/*
-Explicitly disable bracketed paste in the interactive interpreter, even if it's
-set in the inputrc, is enabled by default (eg GNU Readline 8.1), or a user calls
-readline.read_init_file(). The Python REPL has not implemented bracketed
-paste support. Also, bracketed mode writes the "\x1b[?2004h" escape sequence
-into stdout which causes test failures in applications that don't support it.
-It can still be explicitly enabled by calling readline.parse_and_bind("set
-enable-bracketed-paste on"). See bpo-42819 for more details.
-
-This should be removed if bracketed paste mode is implemented (bpo-39820).
-*/
-
-static void
-disable_bracketed_paste(void)
-{
-    if (!using_libedit_emulation) {
-        rl_variable_bind ("enable-bracketed-paste", "off");
-    }
-}
-
 /* Exported function to send one line to readline's init file parser */
 
 static PyObject *
-parse_and_bind(PyObject *self, PyObject *string)
+parse_and_bind(PyObject *self, PyObject *args)
 {
-    char *copy;
-    PyObject *encoded = encode(string);
-    if (encoded == NULL) {
+    char *s, *copy;
+    if (!PyArg_ParseTuple(args, "s:parse_and_bind", &s))
         return NULL;
-    }
     /* Make a copy -- rl_parse_and_bind() modifies its argument */
     /* Bernard Herzog */
-    copy = PyMem_Malloc(1 + PyBytes_GET_SIZE(encoded));
-    if (copy == NULL) {
-        Py_DECREF(encoded);
+    copy = malloc(1 + strlen(s));
+    if (copy == NULL)
         return PyErr_NoMemory();
-    }
-    strcpy(copy, PyBytes_AS_STRING(encoded));
-    Py_DECREF(encoded);
+    strcpy(copy, s);
     rl_parse_and_bind(copy);
-    PyMem_Free(copy); /* Free the copy */
+    free(copy); /* Free the copy */
     Py_RETURN_NONE;
 }
 
 PyDoc_STRVAR(doc_parse_and_bind,
 "parse_and_bind(string) -> None\n\
-Execute the init line provided in the string argument.");
+Parse and execute single line of a readline init file.");
 
 
 /* Exported function to parse a readline init file */
@@ -206,14 +119,13 @@ read_init_file(PyObject *self, PyObject *args)
     } else
         errno = rl_read_init_file(NULL);
     if (errno)
-        return PyErr_SetFromErrno(PyExc_OSError);
-    disable_bracketed_paste();
+        return PyErr_SetFromErrno(PyExc_IOError);
     Py_RETURN_NONE;
 }
 
 PyDoc_STRVAR(doc_read_init_file,
 "read_init_file([filename]) -> None\n\
-Execute a readline initialization file.\n\
+Parse a readline initialization file.\n\
 The default filename is the last filename used.");
 
 
@@ -233,7 +145,7 @@ read_history_file(PyObject *self, PyObject *args)
     } else
         errno = read_history(NULL);
     if (errno)
-        return PyErr_SetFromErrno(PyExc_OSError);
+        return PyErr_SetFromErrno(PyExc_IOError);
     Py_RETURN_NONE;
 }
 
@@ -268,7 +180,7 @@ write_history_file(PyObject *self, PyObject *args)
     Py_XDECREF(filename_bytes);
     errno = err;
     if (errno)
-        return PyErr_SetFromErrno(PyExc_OSError);
+        return PyErr_SetFromErrno(PyExc_IOError);
     Py_RETURN_NONE;
 }
 
@@ -278,43 +190,6 @@ Save a readline history file.\n\
 The default filename is ~/.history.");
 
 
-#ifdef HAVE_RL_APPEND_HISTORY
-/* Exported function to save part of a readline history file */
-
-static PyObject *
-append_history_file(PyObject *self, PyObject *args)
-{
-    int nelements;
-    PyObject *filename_obj = Py_None, *filename_bytes;
-    char *filename;
-    int err;
-    if (!PyArg_ParseTuple(args, "i|O:append_history_file", &nelements, &filename_obj))
-        return NULL;
-    if (filename_obj != Py_None) {
-        if (!PyUnicode_FSConverter(filename_obj, &filename_bytes))
-            return NULL;
-        filename = PyBytes_AsString(filename_bytes);
-    } else {
-        filename_bytes = NULL;
-        filename = NULL;
-    }
-    errno = err = append_history(nelements, filename);
-    if (!err && _history_length >= 0)
-        history_truncate_file(filename, _history_length);
-    Py_XDECREF(filename_bytes);
-    errno = err;
-    if (errno)
-        return PyErr_SetFromErrno(PyExc_OSError);
-    Py_RETURN_NONE;
-}
-
-PyDoc_STRVAR(doc_append_history_file,
-"append_history_file(nelements[, filename]) -> None\n\
-Append the last nelements items of the history list to file.\n\
-The default filename is ~/.history.");
-#endif
-
-
 /* Set history length */
 
 static PyObject*
@@ -329,7 +204,7 @@ set_history_length(PyObject *self, PyObject *args)
 
 PyDoc_STRVAR(set_history_length_doc,
 "set_history_length(length) -> None\n\
-set the maximal number of lines which will be written to\n\
+set the maximal number of items which will be written to\n\
 the history file. A negative length is used to inhibit\n\
 history truncation.");
 
@@ -344,7 +219,7 @@ get_history_length(PyObject *self, PyObject *noarg)
 
 PyDoc_STRVAR(get_history_length_doc,
 "get_history_length() -> int\n\
-return the maximum number of lines that will be written to\n\
+return the maximum number of items that will be written to\n\
 the history file.");
 
 
@@ -362,8 +237,10 @@ set_hook(const char *funcname, PyObject **hook_var, PyObject *args)
         Py_CLEAR(*hook_var);
     }
     else if (PyCallable_Check(function)) {
+        PyObject *tmp = *hook_var;
         Py_INCREF(function);
-        Py_XSETREF(*hook_var, function);
+        *hook_var = function;
+        Py_XDECREF(tmp);
     }
     else {
         PyErr_Format(PyExc_TypeError,
@@ -375,16 +252,25 @@ set_hook(const char *funcname, PyObject **hook_var, PyObject *args)
 }
 
 
+/* Exported functions to specify hook functions in Python */
+
+static PyObject *completion_display_matches_hook = NULL;
+static PyObject *startup_hook = NULL;
+
+#ifdef HAVE_RL_PRE_INPUT_HOOK
+static PyObject *pre_input_hook = NULL;
+#endif
+
 static PyObject *
 set_completion_display_matches_hook(PyObject *self, PyObject *args)
 {
     PyObject *result = set_hook("completion_display_matches_hook",
-                    &readlinestate_global->completion_display_matches_hook, args);
+                    &completion_display_matches_hook, args);
 #ifdef HAVE_RL_COMPLETION_DISPLAY_MATCHES_HOOK
     /* We cannot set this hook globally, since it replaces the
        default completion display. */
     rl_completion_display_matches_hook =
-        readlinestate_global->completion_display_matches_hook ?
+        completion_display_matches_hook ?
 #if defined(_RL_FUNCTION_TYPEDEF)
         (rl_compdisp_func_t *)on_completion_display_matches_hook : 0;
 #else
@@ -405,12 +291,12 @@ once each time matches need to be displayed.");
 static PyObject *
 set_startup_hook(PyObject *self, PyObject *args)
 {
-    return set_hook("startup_hook", &readlinestate_global->startup_hook, args);
+    return set_hook("startup_hook", &startup_hook, args);
 }
 
 PyDoc_STRVAR(doc_set_startup_hook,
 "set_startup_hook([function]) -> None\n\
-Set or remove the function invoked by the rl_startup_hook callback.\n\
+Set or remove the startup_hook function.\n\
 The function is called with no arguments just\n\
 before readline prints the first prompt.");
 
@@ -422,12 +308,12 @@ before readline prints the first prompt.");
 static PyObject *
 set_pre_input_hook(PyObject *self, PyObject *args)
 {
-    return set_hook("pre_input_hook", &readlinestate_global->pre_input_hook, args);
+    return set_hook("pre_input_hook", &pre_input_hook, args);
 }
 
 PyDoc_STRVAR(doc_set_pre_input_hook,
 "set_pre_input_hook([function]) -> None\n\
-Set or remove the function invoked by the rl_pre_input_hook callback.\n\
+Set or remove the pre_input_hook function.\n\
 The function is called with no arguments after the first prompt\n\
 has been printed and just before readline starts reading input\n\
 characters.");
@@ -435,6 +321,14 @@ characters.");
 #endif
 
 
+/* Exported function to specify a word completer in Python */
+
+static PyObject *completer = NULL;
+
+static PyObject *begidx = NULL;
+static PyObject *endidx = NULL;
+
+
 /* Get the completion type for the scope of the tab-completion */
 static PyObject *
 get_completion_type(PyObject *self, PyObject *noarg)
@@ -452,13 +346,13 @@ Get the type of completion being attempted.");
 static PyObject *
 get_begidx(PyObject *self, PyObject *noarg)
 {
-    Py_INCREF(readlinestate_global->begidx);
-    return readlinestate_global->begidx;
+    Py_INCREF(begidx);
+    return begidx;
 }
 
 PyDoc_STRVAR(doc_get_begidx,
 "get_begidx() -> int\n\
-get the beginning index of the completion scope");
+get the beginning index of the readline tab-completion scope");
 
 
 /* Get the ending index for the scope of the tab-completion */
@@ -466,34 +360,32 @@ get the beginning index of the completion scope");
 static PyObject *
 get_endidx(PyObject *self, PyObject *noarg)
 {
-    Py_INCREF(readlinestate_global->endidx);
-    return readlinestate_global->endidx;
+    Py_INCREF(endidx);
+    return endidx;
 }
 
 PyDoc_STRVAR(doc_get_endidx,
 "get_endidx() -> int\n\
-get the ending index of the completion scope");
+get the ending index of the readline tab-completion scope");
 
 
 /* Set the tab-completion word-delimiters that readline uses */
 
 static PyObject *
-set_completer_delims(PyObject *self, PyObject *string)
+set_completer_delims(PyObject *self, PyObject *args)
 {
     char *break_chars;
-    PyObject *encoded = encode(string);
-    if (encoded == NULL) {
+
+    if (!PyArg_ParseTuple(args, "s:set_completer_delims", &break_chars)) {
         return NULL;
     }
     /* Keep a reference to the allocated memory in the module state in case
        some other module modifies rl_completer_word_break_characters
        (see issue #17289). */
-    break_chars = strdup(PyBytes_AS_STRING(encoded));
-    Py_DECREF(encoded);
-    if (break_chars) {
-        free(completer_word_break_characters);
-        completer_word_break_characters = break_chars;
-        rl_completer_word_break_characters = break_chars;
+    free(completer_word_break_characters);
+    completer_word_break_characters = strdup(break_chars);
+    if (completer_word_break_characters) {
+        rl_completer_word_break_characters = completer_word_break_characters;
         Py_RETURN_NONE;
     }
     else
@@ -502,7 +394,7 @@ set_completer_delims(PyObject *self, PyObject *string)
 
 PyDoc_STRVAR(doc_set_completer_delims,
 "set_completer_delims(string) -> None\n\
-set the word delimiters for completion");
+set the readline word delimiters for tab-completion");
 
 /* _py_free_history_entry: Utility function to free a history entry. */
 
@@ -542,7 +434,7 @@ py_remove_history(PyObject *self, PyObject *args)
     int entry_number;
     HIST_ENTRY *entry;
 
-    if (!PyArg_ParseTuple(args, "i:remove_history_item", &entry_number))
+    if (!PyArg_ParseTuple(args, "i:remove_history", &entry_number))
         return NULL;
     if (entry_number < 0) {
         PyErr_SetString(PyExc_ValueError,
@@ -569,11 +461,10 @@ static PyObject *
 py_replace_history(PyObject *self, PyObject *args)
 {
     int entry_number;
-    PyObject *line;
-    PyObject *encoded;
+    char *line;
     HIST_ENTRY *old_entry;
 
-    if (!PyArg_ParseTuple(args, "iU:replace_history_item", &entry_number,
+    if (!PyArg_ParseTuple(args, "is:replace_history", &entry_number,
                           &line)) {
         return NULL;
     }
@@ -582,12 +473,7 @@ py_replace_history(PyObject *self, PyObject *args)
                         "History index cannot be negative");
         return NULL;
     }
-    encoded = encode(line);
-    if (encoded == NULL) {
-        return NULL;
-    }
-    old_entry = replace_history_entry(entry_number, PyBytes_AS_STRING(encoded), (void *)NULL);
-    Py_DECREF(encoded);
+    old_entry = replace_history_entry(entry_number, line, (void *)NULL);
     if (!old_entry) {
         PyErr_Format(PyExc_ValueError,
                      "No history item at position %d",
@@ -606,38 +492,20 @@ replaces history item given by its position with contents of line");
 /* Add a line to the history buffer */
 
 static PyObject *
-py_add_history(PyObject *self, PyObject *string)
+py_add_history(PyObject *self, PyObject *args)
 {
-    PyObject *encoded = encode(string);
-    if (encoded == NULL) {
+    char *line;
+
+    if(!PyArg_ParseTuple(args, "s:add_history", &line)) {
         return NULL;
     }
-    add_history(PyBytes_AS_STRING(encoded));
-    Py_DECREF(encoded);
+    add_history(line);
     Py_RETURN_NONE;
 }
 
 PyDoc_STRVAR(doc_add_history,
 "add_history(string) -> None\n\
-add an item to the history buffer");
-
-static int should_auto_add_history = 1;
-
-/* Enable or disable automatic history */
-
-static PyObject *
-py_set_auto_history(PyObject *self, PyObject *args)
-{
-    if (!PyArg_ParseTuple(args, "p:set_auto_history",
-                          &should_auto_add_history)) {
-        return NULL;
-    }
-    Py_RETURN_NONE;
-}
-
-PyDoc_STRVAR(doc_set_auto_history,
-"set_auto_history(enabled) -> None\n\
-Enables or disables automatic history.");
+add a line to the history buffer");
 
 
 /* Get the tab-completion word-delimiters that readline uses */
@@ -645,12 +513,12 @@ Enables or disables automatic history.");
 static PyObject *
 get_completer_delims(PyObject *self, PyObject *noarg)
 {
-    return decode(rl_completer_word_break_characters);
+    return PyUnicode_FromString(rl_completer_word_break_characters);
 }
 
 PyDoc_STRVAR(doc_get_completer_delims,
 "get_completer_delims() -> string\n\
-get the word delimiters for completion");
+get the readline word delimiters for tab-completion");
 
 
 /* Set the completer function */
@@ -658,7 +526,7 @@ get the word delimiters for completion");
 static PyObject *
 set_completer(PyObject *self, PyObject *args)
 {
-    return set_hook("completer", &readlinestate_global->completer, args);
+    return set_hook("completer", &completer, args);
 }
 
 PyDoc_STRVAR(doc_set_completer,
@@ -672,11 +540,11 @@ It should return the next possible completion starting with 'text'.");
 static PyObject *
 get_completer(PyObject *self, PyObject *noargs)
 {
-    if (readlinestate_global->completer == NULL) {
+    if (completer == NULL) {
         Py_RETURN_NONE;
     }
-    Py_INCREF(readlinestate_global->completer);
-    return readlinestate_global->completer;
+    Py_INCREF(completer);
+    return completer;
 }
 
 PyDoc_STRVAR(doc_get_completer,
@@ -711,8 +579,9 @@ get_history_item(PyObject *self, PyObject *args)
     int idx = 0;
     HIST_ENTRY *hist_ent;
 
-    if (!PyArg_ParseTuple(args, "i:get_history_item", &idx))
+    if (!PyArg_ParseTuple(args, "i:index", &idx))
         return NULL;
+#ifdef  __APPLE__
     if (using_libedit_emulation) {
         /* Older versions of libedit's readline emulation
          * use 0-based indexes, while readline and newer
@@ -732,8 +601,9 @@ get_history_item(PyObject *self, PyObject *args)
             Py_RETURN_NONE;
         }
     }
+#endif /* __APPLE__ */
     if ((hist_ent = history_get(idx)))
-        return decode(hist_ent->line);
+        return PyUnicode_FromString(hist_ent->line);
     else {
         Py_RETURN_NONE;
     }
@@ -762,7 +632,7 @@ return the current (not the maximum) length of history.");
 static PyObject *
 get_line_buffer(PyObject *self, PyObject *noarg)
 {
-    return decode(rl_line_buffer);
+    return PyUnicode_FromString(rl_line_buffer);
 }
 
 PyDoc_STRVAR(doc_get_line_buffer,
@@ -790,20 +660,18 @@ Clear the current readline history.");
 /* Exported function to insert text into the line buffer */
 
 static PyObject *
-insert_text(PyObject *self, PyObject *string)
+insert_text(PyObject *self, PyObject *args)
 {
-    PyObject *encoded = encode(string);
-    if (encoded == NULL) {
+    char *s;
+    if (!PyArg_ParseTuple(args, "s:insert_text", &s))
         return NULL;
-    }
-    rl_insert_text(PyBytes_AS_STRING(encoded));
-    Py_DECREF(encoded);
+    rl_insert_text(s);
     Py_RETURN_NONE;
 }
 
 PyDoc_STRVAR(doc_insert_text,
 "insert_text(string) -> None\n\
-Insert text into the line buffer at the cursor position.");
+Insert text into the command line.");
 
 
 /* Redisplay the line buffer */
@@ -825,19 +693,15 @@ contents of the line buffer.");
 
 static struct PyMethodDef readline_methods[] =
 {
-    {"parse_and_bind", parse_and_bind, METH_O, doc_parse_and_bind},
+    {"parse_and_bind", parse_and_bind, METH_VARARGS, doc_parse_and_bind},
     {"get_line_buffer", get_line_buffer, METH_NOARGS, doc_get_line_buffer},
-    {"insert_text", insert_text, METH_O, doc_insert_text},
+    {"insert_text", insert_text, METH_VARARGS, doc_insert_text},
     {"redisplay", redisplay, METH_NOARGS, doc_redisplay},
     {"read_init_file", read_init_file, METH_VARARGS, doc_read_init_file},
     {"read_history_file", read_history_file,
      METH_VARARGS, doc_read_history_file},
     {"write_history_file", write_history_file,
      METH_VARARGS, doc_write_history_file},
-#ifdef HAVE_RL_APPEND_HISTORY
-    {"append_history_file", append_history_file,
-     METH_VARARGS, doc_append_history_file},
-#endif
     {"get_history_item", get_history_item,
      METH_VARARGS, doc_get_history_item},
     {"get_current_history_length", (PyCFunction)get_current_history_length,
@@ -854,9 +718,8 @@ static struct PyMethodDef readline_methods[] =
     {"get_endidx", get_endidx, METH_NOARGS, doc_get_endidx},
 
     {"set_completer_delims", set_completer_delims,
-     METH_O, doc_set_completer_delims},
-    {"set_auto_history", py_set_auto_history, METH_VARARGS, doc_set_auto_history},
-    {"add_history", py_add_history, METH_O, doc_add_history},
+     METH_VARARGS, doc_set_completer_delims},
+    {"add_history", py_add_history, METH_VARARGS, doc_add_history},
     {"remove_history_item", py_remove_history, METH_VARARGS, doc_remove_history},
     {"replace_history_item", py_replace_history, METH_VARARGS, doc_replace_history},
     {"get_completer_delims", get_completer_delims,
@@ -885,13 +748,16 @@ on_hook(PyObject *func)
     int result = 0;
     if (func != NULL) {
         PyObject *r;
-        r = _PyObject_CallNoArg(func);
+#ifdef WITH_THREAD
+        PyGILState_STATE gilstate = PyGILState_Ensure();
+#endif
+        r = PyObject_CallFunction(func, NULL);
         if (r == NULL)
             goto error;
         if (r == Py_None)
             result = 0;
         else {
-            result = _PyLong_AsInt(r);
+            result = PyLong_AsLong(r);
             if (result == -1 && PyErr_Occurred())
                 goto error;
         }
@@ -901,6 +767,9 @@ on_hook(PyObject *func)
         PyErr_Clear();
         Py_XDECREF(r);
       done:
+#ifdef WITH_THREAD
+        PyGILState_Release(gilstate);
+#endif
         return result;
     }
     return result;
@@ -913,11 +782,7 @@ on_startup_hook(void)
 on_startup_hook()
 #endif
 {
-    int r;
-    PyGILState_STATE gilstate = PyGILState_Ensure();
-    r = on_hook(readlinestate_global->startup_hook);
-    PyGILState_Release(gilstate);
-    return r;
+    return on_hook(startup_hook);
 }
 
 #ifdef HAVE_RL_PRE_INPUT_HOOK
@@ -928,11 +793,7 @@ on_pre_input_hook(void)
 on_pre_input_hook()
 #endif
 {
-    int r;
-    PyGILState_STATE gilstate = PyGILState_Ensure();
-    r = on_hook(readlinestate_global->pre_input_hook);
-    PyGILState_Release(gilstate);
-    return r;
+    return on_hook(pre_input_hook);
 }
 #endif
 
@@ -945,22 +806,24 @@ on_completion_display_matches_hook(char **matches,
                                    int num_matches, int max_length)
 {
     int i;
-    PyObject *sub, *m=NULL, *s=NULL, *r=NULL;
+    PyObject *m=NULL, *s=NULL, *r=NULL;
+#ifdef WITH_THREAD
     PyGILState_STATE gilstate = PyGILState_Ensure();
+#endif
     m = PyList_New(num_matches);
     if (m == NULL)
         goto error;
     for (i = 0; i < num_matches; i++) {
-        s = decode(matches[i+1]);
+        s = PyUnicode_FromString(matches[i+1]);
         if (s == NULL)
             goto error;
-        PyList_SET_ITEM(m, i, s);
+        if (PyList_SetItem(m, i, s) == -1)
+            goto error;
     }
-    sub = decode(matches[0]);
-    r = PyObject_CallFunction(readlinestate_global->completion_display_matches_hook,
-                              "NNi", sub, m, max_length);
+    r = PyObject_CallFunction(completion_display_matches_hook,
+                              "sOi", matches[0], m, max_length);
 
-    m=NULL;
+    Py_DECREF(m); m=NULL;
 
     if (r == NULL ||
         (r != Py_None && PyLong_AsLong(r) == -1 && PyErr_Occurred())) {
@@ -974,29 +837,11 @@ on_completion_display_matches_hook(char **matches,
         Py_XDECREF(m);
         Py_XDECREF(r);
     }
+#ifdef WITH_THREAD
     PyGILState_Release(gilstate);
-}
-
-#endif
-
-#ifdef HAVE_RL_RESIZE_TERMINAL
-static volatile sig_atomic_t sigwinch_received;
-static PyOS_sighandler_t sigwinch_ohandler;
-
-static void
-readline_sigwinch_handler(int signum)
-{
-    sigwinch_received = 1;
-    if (sigwinch_ohandler &&
-            sigwinch_ohandler != SIG_IGN && sigwinch_ohandler != SIG_DFL)
-        sigwinch_ohandler(signum);
-
-#ifndef HAVE_SIGACTION
-    /* If the handler was installed with signal() rather than sigaction(),
-    we need to reinstall it. */
-    PyOS_setsig(SIGWINCH, readline_sigwinch_handler);
 #endif
 }
+
 #endif
 
 /* C function to call the Python completer. */
@@ -1005,23 +850,23 @@ static char *
 on_completion(const char *text, int state)
 {
     char *result = NULL;
-    if (readlinestate_global->completer != NULL) {
-        PyObject *r = NULL, *t;
+    if (completer != NULL) {
+        PyObject *r;
+#ifdef WITH_THREAD
         PyGILState_STATE gilstate = PyGILState_Ensure();
+#endif
         rl_attempted_completion_over = 1;
-        t = decode(text);
-        r = PyObject_CallFunction(readlinestate_global->completer, "Ni", t, state);
+        r = PyObject_CallFunction(completer, "si", text, state);
         if (r == NULL)
             goto error;
         if (r == Py_None) {
             result = NULL;
         }
         else {
-            PyObject *encoded = encode(r);
-            if (encoded == NULL)
+            char *s = _PyUnicode_AsString(r);
+            if (s == NULL)
                 goto error;
-            result = strdup(PyBytes_AS_STRING(encoded));
-            Py_DECREF(encoded);
+            result = strdup(s);
         }
         Py_DECREF(r);
         goto done;
@@ -1029,7 +874,9 @@ on_completion(const char *text, int state)
         PyErr_Clear();
         Py_XDECREF(r);
       done:
+#ifdef WITH_THREAD
         PyGILState_Release(gilstate);
+#endif
         return result;
     }
     return result;
@@ -1042,52 +889,24 @@ on_completion(const char *text, int state)
 static char **
 flex_complete(const char *text, int start, int end)
 {
-    char **result;
-    char saved;
-    size_t start_size, end_size;
-    wchar_t *s;
-    PyGILState_STATE gilstate = PyGILState_Ensure();
 #ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
     rl_completion_append_character ='\0';
 #endif
 #ifdef HAVE_RL_COMPLETION_SUPPRESS_APPEND
     rl_completion_suppress_append = 0;
 #endif
-
-    saved = rl_line_buffer[start];
-    rl_line_buffer[start] = 0;
-    s = Py_DecodeLocale(rl_line_buffer, &start_size);
-    rl_line_buffer[start] = saved;
-    if (s == NULL) {
-        goto done;
-    }
-    PyMem_RawFree(s);
-    saved = rl_line_buffer[end];
-    rl_line_buffer[end] = 0;
-    s = Py_DecodeLocale(rl_line_buffer + start, &end_size);
-    rl_line_buffer[end] = saved;
-    if (s == NULL) {
-        goto done;
-    }
-    PyMem_RawFree(s);
-    start = (int)start_size;
-    end = start + (int)end_size;
-
-done:
-    Py_XDECREF(readlinestate_global->begidx);
-    Py_XDECREF(readlinestate_global->endidx);
-    readlinestate_global->begidx = PyLong_FromLong((long) start);
-    readlinestate_global->endidx = PyLong_FromLong((long) end);
-    result = completion_matches((char *)text, *on_completion);
-    PyGILState_Release(gilstate);
-    return result;
+    Py_XDECREF(begidx);
+    Py_XDECREF(endidx);
+    begidx = PyLong_FromLong((long) start);
+    endidx = PyLong_FromLong((long) end);
+    return completion_matches(text, *on_completion);
 }
 
 
 /* Helper to initialize GNU readline properly. */
 
 static void
-setup_readline(readlinestate *mod_state)
+setup_readline(void)
 {
 #ifdef SAVE_LOCALE
     char *saved_locale = strdup(setlocale(LC_CTYPE, NULL));
@@ -1095,9 +914,7 @@ setup_readline(readlinestate *mod_state)
         Py_FatalError("not enough memory to save locale");
 #endif
 
-    /* The name must be defined before initialization */
-    rl_readline_name = "python";
-
+#ifdef __APPLE__
     /* the libedit readline emulation resets key bindings etc
      * when calling rl_initialize.  So call it upfront
      */
@@ -1114,18 +931,20 @@ setup_readline(readlinestate *mod_state)
         libedit_history_start = 1;
     }
     clear_history();
+#endif /* __APPLE__ */
 
     using_history();
 
+    rl_readline_name = "python";
+#if defined(PYOS_OS2) && defined(PYCC_GCC)
+    /* Allow $if term= in .inputrc to work */
+    rl_terminal_name = getenv("TERM");
+#endif
     /* Force rebind of TAB to insert-tab */
     rl_bind_key('\t', rl_insert);
     /* Bind both ESC-TAB and ESC-ESC to the completion function */
     rl_bind_key_in_map ('\t', rl_complete, emacs_meta_keymap);
     rl_bind_key_in_map ('\033', rl_complete, emacs_meta_keymap);
-#ifdef HAVE_RL_RESIZE_TERMINAL
-    /* Set up signal handler for window resize */
-    sigwinch_ohandler = PyOS_setsig(SIGWINCH, readline_sigwinch_handler);
-#endif
     /* Set our hook functions */
     rl_startup_hook = on_startup_hook;
 #ifdef HAVE_RL_PRE_INPUT_HOOK
@@ -1139,42 +958,29 @@ setup_readline(readlinestate *mod_state)
         strdup(" \t\n`~!@#$%^&*()-=+[{]}\\|;:'\",<>/?");
         /* All nonalphanums except '.' */
 
-    mod_state->begidx = PyLong_FromLong(0L);
-    mod_state->endidx = PyLong_FromLong(0L);
-
-    if (!using_libedit_emulation)
-    {
-        if (!isatty(STDOUT_FILENO)) {
-            /* Issue #19884: stdout is not a terminal. Disable meta modifier
-               keys to not write the ANSI sequence "\033[1034h" into stdout. On
-               terminals supporting 8 bit characters like TERM=xterm-256color
-               (which is now the default Fedora since Fedora 18), the meta key is
-               used to enable support of 8 bit characters (ANSI sequence
-               "\033[1034h").
-
-               With libedit, this call makes readline() crash. */
-            rl_variable_bind ("enable-meta-key", "off");
-        }
-    }
-
+    begidx = PyLong_FromLong(0L);
+    endidx = PyLong_FromLong(0L);
     /* Initialize (allows .inputrc to override)
      *
      * XXX: A bug in the readline-2.2 library causes a memory leak
      * inside this function.  Nothing we can do about it.
      */
+#ifdef __APPLE__
     if (using_libedit_emulation)
         rl_read_init_file(NULL);
     else
+#endif /* __APPLE__ */
         rl_initialize();
 
-    disable_bracketed_paste();
-
     RESTORE_LOCALE(saved_locale)
 }
 
 /* Wrapper around GNU readline that handles signals differently. */
 
-static char *completed_input_string;
+
+#if defined(HAVE_RL_CALLBACK) && defined(HAVE_SELECT)
+
+static  char *completed_input_string;
 static void
 rlhandler(char *text)
 {
@@ -1182,8 +988,10 @@ rlhandler(char *text)
     rl_callback_handler_remove();
 }
 
+extern PyThreadState* _PyOS_ReadlineTState;
+
 static char *
-readline_until_enter_or_signal(const char *prompt, int *signal)
+readline_until_enter_or_signal(char *prompt, int *signal)
 {
     char * not_done_reading = "";
     fd_set selectset;
@@ -1209,13 +1017,6 @@ readline_until_enter_or_signal(const char *prompt, int *signal)
             struct timeval *timeoutp = NULL;
             if (PyOS_InputHook)
                 timeoutp = &timeout;
-#ifdef HAVE_RL_RESIZE_TERMINAL
-            /* Update readline's view of the window size after SIGWINCH */
-            if (sigwinch_received) {
-                sigwinch_received = 0;
-                rl_resize_terminal();
-            }
-#endif
             FD_SET(fileno(rl_instream), &selectset);
             /* select resets selectset if no input was available */
             has_input = select(fileno(rl_instream) + 1, &selectset,
@@ -1229,14 +1030,15 @@ readline_until_enter_or_signal(const char *prompt, int *signal)
         }
         else if (err == EINTR) {
             int s;
+#ifdef WITH_THREAD
             PyEval_RestoreThread(_PyOS_ReadlineTState);
+#endif
             s = PyErr_CheckSignals();
+#ifdef WITH_THREAD
             PyEval_SaveThread();
+#endif
             if (s < 0) {
                 rl_free_line_state();
-#if defined(RL_READLINE_VERSION) && RL_READLINE_VERSION >= 0x0700
-                rl_callback_sigcleanup();
-#endif
                 rl_cleanup_after_signal();
                 rl_callback_handler_remove();
                 *signal = 1;
@@ -1249,18 +1051,59 @@ readline_until_enter_or_signal(const char *prompt, int *signal)
 }
 
 
+#else
+
+/* Interrupt handler */
+
+static jmp_buf jbuf;
+
+/* ARGSUSED */
+static void
+onintr(int sig)
+{
+    longjmp(jbuf, 1);
+}
+
+
 static char *
-call_readline(FILE *sys_stdin, FILE *sys_stdout, const char *prompt)
+readline_until_enter_or_signal(char *prompt, int *signal)
+{
+    PyOS_sighandler_t old_inthandler;
+    char *p;
+
+    *signal = 0;
+
+    old_inthandler = PyOS_setsig(SIGINT, onintr);
+    if (setjmp(jbuf)) {
+#ifdef HAVE_SIGRELSE
+        /* This seems necessary on SunOS 4.1 (Rasmus Hahn) */
+        sigrelse(SIGINT);
+#endif
+        PyOS_setsig(SIGINT, old_inthandler);
+        *signal = 1;
+        return NULL;
+    }
+    rl_event_hook = PyOS_InputHook;
+    p = readline(prompt);
+    PyOS_setsig(SIGINT, old_inthandler);
+
+    return p;
+}
+#endif /*defined(HAVE_RL_CALLBACK) && defined(HAVE_SELECT) */
+
+
+static char *
+call_readline(FILE *sys_stdin, FILE *sys_stdout, char *prompt)
 {
     size_t n;
-    char *p;
+    char *p, *q;
     int signal;
 
 #ifdef SAVE_LOCALE
     char *saved_locale = strdup(setlocale(LC_CTYPE, NULL));
     if (!saved_locale)
         Py_FatalError("not enough memory to save locale");
-    _Py_SetLocaleFromEnv(LC_CTYPE);
+    setlocale(LC_CTYPE, "");
 #endif
 
     if (sys_stdin != rl_instream || sys_stdout != rl_outstream) {
@@ -1279,9 +1122,9 @@ call_readline(FILE *sys_stdin, FILE *sys_stdout, const char *prompt)
         return NULL;
     }
 
-    /* We got an EOF, return an empty string. */
+    /* We got an EOF, return a empty string. */
     if (p == NULL) {
-        p = PyMem_RawMalloc(1);
+        p = PyMem_Malloc(1);
         if (p != NULL)
             *p = '\0';
         RESTORE_LOCALE(saved_locale)
@@ -1290,28 +1133,28 @@ call_readline(FILE *sys_stdin, FILE *sys_stdout, const char *prompt)
 
     /* we have a valid line */
     n = strlen(p);
-    if (should_auto_add_history && n > 0) {
+    if (n > 0) {
         const char *line;
         int length = _py_get_history_length();
-        if (length > 0) {
-            HIST_ENTRY *hist_ent;
+        if (length > 0)
+#ifdef __APPLE__
             if (using_libedit_emulation) {
                 /* handle older 0-based or newer 1-based indexing */
-                hist_ent = history_get(length + libedit_history_start - 1);
+                line = (const char *)history_get(length + libedit_history_start - 1)->line;
             } else
-                hist_ent = history_get(length);
-            line = hist_ent ? hist_ent->line : "";
-        } else
+#endif /* __APPLE__ */
+            line = (const char *)history_get(length)->line;
+        else
             line = "";
         if (strcmp(p, line))
             add_history(p);
     }
     /* Copy the malloc'ed buffer into a PyMem_Malloc'ed one and
        release the original. */
-    char *q = p;
-    p = PyMem_RawMalloc(n+2);
+    q = p;
+    p = PyMem_Malloc(n+2);
     if (p != NULL) {
-        memcpy(p, q, n);
+        strncpy(p, q, n);
         p[n] = '\n';
         p[n+1] = '\0';
     }
@@ -1326,19 +1169,21 @@ call_readline(FILE *sys_stdin, FILE *sys_stdout, const char *prompt)
 PyDoc_STRVAR(doc_module,
 "Importing this module enables command line editing using GNU readline.");
 
+#ifdef __APPLE__
 PyDoc_STRVAR(doc_module_le,
 "Importing this module enables command line editing using libedit readline.");
+#endif /* __APPLE__ */
 
 static struct PyModuleDef readlinemodule = {
     PyModuleDef_HEAD_INIT,
     "gnureadline",
     doc_module,
-    sizeof(readlinestate),
+    -1,
     readline_methods,
     NULL,
-    readline_traverse,
-    readline_clear,
-    readline_free
+    NULL,
+    NULL,
+    NULL
 };
 
 
@@ -1346,8 +1191,8 @@ PyMODINIT_FUNC
 PyInit_gnureadline(void)
 {
     PyObject *m;
-    readlinestate *mod_state;
 
+#ifdef __APPLE__
     if (strncmp(rl_library_version, libedit_version_tag, strlen(libedit_version_tag)) == 0) {
         using_libedit_emulation = 1;
     }
@@ -1355,33 +1200,14 @@ PyInit_gnureadline(void)
     if (using_libedit_emulation)
         readlinemodule.m_doc = doc_module_le;
 
+#endif /* __APPLE__ */
 
     m = PyModule_Create(&readlinemodule);
 
     if (m == NULL)
         return NULL;
 
-    if (PyModule_AddIntConstant(m, "_READLINE_VERSION",
-                                RL_READLINE_VERSION) < 0) {
-        goto error;
-    }
-    if (PyModule_AddIntConstant(m, "_READLINE_RUNTIME_VERSION",
-                                rl_readline_version) < 0) {
-        goto error;
-    }
-    if (PyModule_AddStringConstant(m, "_READLINE_LIBRARY_VERSION",
-                                   rl_library_version) < 0)
-    {
-        goto error;
-    }
-
-    mod_state = (readlinestate *) PyModule_GetState(m);
     PyOS_ReadlineFunctionPointer = call_readline;
-    setup_readline(mod_state);
-
+    setup_readline();
     return m;
-
-error:
-    Py_DECREF(m);
-    return NULL;
 }
